/*
 * Copyright (c) 1991-2023 by STEP Tools Inc. 
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and
 * its documentation is hereby granted, provided that this copyright
 * notice and license appear on all copies of the software.
 * 
 * STEP TOOLS MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE
 * SUITABILITY OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING
 * BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT. STEP TOOLS
 * SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A
 * RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS
 * DERIVATIVES.
 *
 * Author: David Loffredo (loffredo@steptools.com)
 */

#ifndef TOOLPATH_H
#define TOOLPATH_H

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

// For floating point comparisons.  Most CNCs will not go to more than
// six digits of precision so 1e-8 would probably be sufficient.
//
#define STEPNC_EPSILON 1e-15

#define STEPNC_FLOAT_EQUAL(a,b)           STEPNC_FLOAT_ZERO((a)-(b))
#define STEPNC_FLOAT_NOT_EQUAL(a,b)       STEPNC_FLOAT_NOT_ZERO((a)-(b))
#define STEPNC_FLOAT_ZERO(a)              (fabs(a) < STEPNC_EPSILON)
#define STEPNC_FLOAT_NOT_ZERO(a)          (fabs(a) > STEPNC_EPSILON)

class ToolpathMaker {
protected:
    RoseDesign * f_design;

    // ------------------------------
    // PREFERRED UNITS
    // A little local enum to keep track of the units that we should
    // use for different quantities.  The file can specify arbitrary
    // units, but normally you use a group of related ones.
    //
    enum unit_system_type {
	inch_deg_ipm_rpm,
	mm_deg_mmpm_rpm
    };

    unit_system_type unit_system;

    // ------------------------------
    // FEEDS AND SPEEDS
    // A little local struct to keep track of machining technology
    // constructs for feeds and speeds so we can avoid duplication
    // when we switch back and forth between a couple of feeds.
    //
    struct techdef {
	techdef * next_tech;
	stp_machining_technology * tech_id;
	double feed;
	double speed;
    };
    int    is_rapid;
    double current_speed;
    double current_feed;
    techdef * current_tech;
    techdef * all_techs;


    // ------------------------------
    // COOLANT STATE
    // A little local struct to keep track of machine functions
    // constructs for coolant so we can avoid duplication when we
    // switch coolant on and off.  Machine functions has other
    // parameters that we do not use at the moment.
    //
    struct mfuncdef {
	mfuncdef * next_mfunc;
	stp_machining_functions * mfunc_id;
	int is_coolant;
	int is_mist;
    };
    int current_coolant;
    int current_mist;
    mfuncdef * current_mfunc;
    mfuncdef * all_mfuncs;

    // ------------------------------
    // CUTTING TOOL
    // In the file, a tool points to all of the operations that use
    // it, so we must save all the cutter definitions until the end,
    // when we know all of the workingsteps.  Keep little local struct
    // with the parameters and a list of using workingsteps.
    //
    struct tooldef {
	tooldef * next_tool;

	double diameter;
	double corner_radius;
	double corner_radius_horizontal;
	double corner_radius_vertical; 
	double tip_outer_angle; 
	double taper_angle; 
	double height;

	unsigned tool_num;
	rose_vector users;
    };
    tooldef * current_tool;
    tooldef * all_tools;


    // ------------------------------
    // LOCATION CURVE STATE
    // Track the points that make up a polyline and the curves that
    // make up segments so that we can write the basis curves, tool
    // axis curves and surface normal curves.  We build the curves at
    // the end of moves so that we can optimize the use of polylines
    // vs composite curves.
    // 
    struct curvedef {
	rose_vector 	points;
	rose_vector 	curves;

	// convenient checking against current values and used to
	// share start/end points between paths.
	double 			last_point[3];  
	stp_cartesian_point * 	last_point_id;
    };

    curvedef loc;
    curvedef axis;

    stp_direction * current_zdir;
    stp_direction * current_xdir;


    // ------------------------------
    // WORKPLANS, OPERATIONS, AND TOOLPATHS
    //
    rose_stack plan_ids;	// nested workplans and positions
    rose_stack plan_idxs;

    unsigned ws_count;
    unsigned tp_count;

    stp_machining_operation * 	current_op;
    stp_machining_toolpath *  	current_tp;
    unsigned 			current_tpidx;

    stp_product_definition * 		current_project_pd;
    stp_product_definition_shape * 	current_workpiece_pds;
    stp_machining_feature_process * 	current_feature_mfp;

    stp_product_context * 		ctx_product;
    stp_product_definition_context * 	ctx_pdef;

public:
    // ------------------------------
    // MAIN PUBLIC INTERFACE
    //
    // Create an instance of the STEPNC writer, then open a file with
    // the Open command.  This will return NULL if the file could not
    // be opened for some reason. Issue a sequence of commands and
    // then Close() the file.  Reset() will also close an open file.
    //
    ToolpathMaker();
    ~ToolpathMaker();

    void Design (RoseDesign * d);
    void Done();
    void Reset();

    // ------------------------------
    // PART AND UNIT INFORMATION
    //
    void PartNo (const char * name);

    void Inches() 	{ unit_system = inch_deg_ipm_rpm; }
    void Millimeters() 	{ unit_system = mm_deg_mmpm_rpm; }

    // ------------------------------
    // SPEEDS, FEEDS AND COOLANT
    //
    // Sets the spindle, feedrate, and coolant state.

    void SpindleSpeed (double speed);

    void Rapid();
    void Feedrate (double speed);

    void CoolantOn();
    void CoolantOff();
    void CoolantMist();

    // ------------------------------
    // TOOL DEFINITIONS
    //
    // Defines a cutter, but does not assign it a tool number.
    void Cutter(
	double diameter, 
	double corner_radius, 
	double corner_radius_horizontal, 
	double corner_radius_vertical, 
	double tip_outer_angle, 
	double taper_angle, 
	double height
	);

    // Loads a particular tool number.  Assigns the tool number to the
    // last defined cutter if the tool number does not already exist.
    //
    void LoadTool (unsigned num);

    // Assigns tool number to last defined cutter
    void SELCTLTool (unsigned num);


    // ------------------------------
    // MOTION COMMANDS
    //
    // Linear move to a given location.   Tool axis is (0,0,1)
    void GoToXYZ(
	const char * label, 
	double x, 
	double y, 
	double z
	);

    // Linear move to a given location and tool axis.
    void GoToXYZ_IJK(
	const char * label, 
	double x, 
	double y, 
	double z,
	double i,
	double j, 
	double k
	);

    // Arc from the current location to a new location, given a
    // center, radius and clockwise/ccw direction.  
    //
    void Arc(
	const char * label, 
	double new_x, 
	double new_y, 
	double new_z, 
	double cx, 
	double cy, 
	double cz, 
	double radius, 
	int ccw
	);

    // Set direction components used to establish the plane that Arc
    // commands use.
    void SetDirection(
	double zi, double zj, double zk, 	// z axis
	double xa, double xb, double xc		// x axis
	);


    // ------------------------------
    // FINE TUNING OF PROGRAM STRUCTURE 
    //
    // By default, a STEP-NC file only has a single workplan, and
    // workingsteps are created as needed to support the requested
    // motions and speeds.  The following functions force additional
    // workingsteps or nested workplans.  
    // 
    // Ending the top level workplan will close the file.
    //
    void NestWorkplan (const char * name);
    void EndWorkplan();

    void Workingstep (const char * name);


    // Various NC functions
    void Connector();
    void GoHome (const char * label);
    void Stop (const char * label);

    void PPrint(
	const char * label, 
	const char * message
	);


    // Unused, from earlier interfaces
    void MultaxOn()	{ /* not needed, automatically handled */ }
    void MultaxOff()	{ /* not needed, automatically handled */ }

protected:
    // ------------------------------
    // Below this line are utility functions used to write specific
    // STEP-NC data entities.  The functions use the make_ prefix and
    // may either write new data objects to the file or return the id
    // of a shared, previously-written object.
    //

    void make_workplan (const char * label);
    void end_workplan();
    void add_to_workplan (stp_machining_process_executable * mpe);

    void make_workingstep (const char * label);
    void end_workingstep();
    void add_to_workingstep (stp_machining_toolpath * tp);

    stp_machining_toolpath * make_toolpath (const char * label);
    stp_machining_feature_process * make_toolpath_feature();

    void start_move(const char * label);
    void end_move();
	
    void make_all_tools();

    stp_machining_technology * make_technology();
    stp_machining_functions *  make_mfunc();

    stp_cartesian_point * make_point (
	const char * label, 
	double x, 
	double y, 
	double z
	);

    stp_curve * make_curve (
	const char * label, 
	curvedef &def
	);

    stp_polyline * make_polyline (
	const char * label,
	rose_vector &pts
	);

    stp_composite_curve * make_composite_curve (
	const char * label,
	rose_vector &curves
	);


    // --------------------
    // Propertyies, measures, and units
    //
    stp_representation_context * make_simple_context();
    stp_representation_context * make_geometry_context();

    stp_representation * make_string_rep (const char * name);

    stp_action_property * make_action_property (
	const char * nm, 
	stp_action_method * am,
	stp_representation * rep
	);

    stp_measure_representation_item * make_length_measure(
	const char * label,
	double val
	);

    stp_measure_representation_item * make_angle_measure(
	const char * label,
	double val
	);


    stp_measure_representation_item * make_feed_measure(
	const char * label,
	double val
	);

    stp_measure_representation_item * make_spindle_measure(
	const char * label,
	double val
	);
    
    stp_named_unit * make_length_unit();
    stp_named_unit * make_mm_unit();
    stp_named_unit * make_inch_unit();

    stp_named_unit * make_angle_unit();
    stp_named_unit * make_radian_unit();
    stp_named_unit * make_degree_unit();

    stp_named_unit * make_steradian_unit();
    stp_named_unit * make_second_unit();
    stp_named_unit * make_minute_unit();
    stp_named_unit * make_revolution_unit();

    stp_derived_unit * make_feed_unit();
    stp_derived_unit * make_mmpm_unit();
    stp_derived_unit * make_ipm_unit();

    stp_derived_unit * make_spindle_unit();
    stp_derived_unit * make_rps_unit();
    stp_derived_unit * make_rpm_unit();

};


#endif /* TOOLPATH_H */
